@@ -1,4 +1,5 @@
 use crate::lex::lex;
+use crate::types::*;
 use crate::SyntaxKind;
 use crate::SyntaxKind::*;
 use crate::DEFAULT_VERSION;
@@ -123,7 +124,7 @@ fn parse(text: &str) -> Parse {
             }
             self.builder.start_node(ENTRY.into());
             self.parse_options_list();
-            for i in 1..3 {
+            for i in 0..4 {
                 if self.current() == Some(NEWLINE) {
                     break;
                 }
@@ -139,18 +140,22 @@ fn parse(text: &str) -> Parse {
                         self.current(),
                         i
                     ));
-                    self.bump();
+                    if self.current().is_some() {
+                        self.bump();
+                    }
                     self.builder.finish_node();
                 } else {
                     self.bump();
                 }
                 self.skip_ws();
             }
-            if self.current() != Some(NEWLINE) {
+            if self.current() != Some(NEWLINE) && self.current() != None {
                 self.builder.start_node(ERROR.into());
                 self.errors
                     .push(format!("expected newline, not {:?}", self.current()));
-                self.bump();
+                if self.current().is_some() {
+                    self.bump();
+                }
                 self.builder.finish_node();
             } else {
                 self.bump();
@@ -163,6 +168,9 @@ fn parse(text: &str) -> Parse {
             if self.current().is_none() {
                 return false;
             }
+            while self.current() == Some(CONTINUATION) {
+                self.bump();
+            }
             if self.current() == Some(WHITESPACE) {
                 return false;
             }
@@ -175,22 +183,25 @@ fn parse(text: &str) -> Parse {
             } else {
                 self.bump();
             }
-            if self.current() != Some(EQUALS) {
+            if self.current() == Some(EQUALS) {
+                self.bump();
+                if self.current() != Some(VALUE) && self.current() != Some(KEY) {
+                    self.builder.start_node(ERROR.into());
+                    self.errors
+                        .push(format!("expected value, got {:?}", self.current()));
+                    self.bump();
+                    self.builder.finish_node();
+                } else {
+                    self.bump();
+                }
+            } else if self.current() == Some(COMMA) {
+            } else {
                 self.builder.start_node(ERROR.into());
                 self.errors.push("expected `=`".to_string());
-                self.bump();
+                if self.current().is_some() {
+                    self.bump();
+                }
                 self.builder.finish_node();
-            } else {
-                self.bump();
-            }
-            if self.current() != Some(VALUE) && self.current() != Some(KEY) {
-                self.builder.start_node(ERROR.into());
-                self.errors
-                    .push(format!("expected value, got {:?}", self.current()));
-                self.bump();
-                self.builder.finish_node();
-            } else {
-                self.bump();
             }
             self.builder.finish_node();
             true
@@ -207,15 +218,35 @@ fn parse(text: &str) -> Parse {
                 if self.current() != Some(EQUALS) {
                     self.builder.start_node(ERROR.into());
                     self.errors.push("expected `=`".to_string());
-                    self.bump();
+                    if self.current().is_some() {
+                        self.bump();
+                    }
                     self.builder.finish_node();
                 } else {
                     self.bump();
                 }
+                let quoted = if self.current() == Some(QUOTE) {
+                    self.bump();
+                    true
+                } else {
+                    false
+                };
                 loop {
+                    if quoted {
+                        if self.current() == Some(QUOTE) {
+                            self.bump();
+                            break;
+                        }
+                        self.skip_ws();
+                    }
                     if !self.parse_option() {
                         break;
                     }
+                    if self.current() == Some(COMMA) {
+                        self.bump();
+                    } else if !quoted {
+                        break;
+                    }
                 }
                 self.builder.finish_node();
                 self.skip_ws();
@@ -229,6 +260,7 @@ fn parse(text: &str) -> Parse {
             if let Some(v) = self.parse_version() {
                 version = v;
             }
+            // TODO: use version to influence parsing
             loop {
                 if !self.parse_watch_entry() {
                     break;
@@ -395,6 +427,141 @@ impl Entry {
         self.0.children().find_map(OptionList::cast)
     }
 
+    pub fn get_option(&self, key: &str) -> Option<String> {
+        self.option_list().and_then(|ol| ol.get_option(key))
+    }
+
+    pub fn has_option(&self, key: &str) -> bool {
+        self.option_list().map_or(false, |ol| ol.has_option(key))
+    }
+
+    /// The name of the secondary source tarball
+    pub fn component(&self) -> Option<String> {
+        self.get_option("component")
+    }
+
+    pub fn ctype(&self) -> Option<ComponentType> {
+        self.get_option("ctype").and_then(|it| it.parse().ok())
+    }
+
+    pub fn compression(&self) -> Option<Compression> {
+        self.get_option("compression")
+            .and_then(|it| it.parse().ok())
+    }
+
+    pub fn repack(&self) -> bool {
+        self.has_option("repack")
+    }
+
+    pub fn repacksuffix(&self) -> Option<String> {
+        self.get_option("repacksuffix")
+    }
+
+    pub fn mode(&self) -> Option<Mode> {
+        self.get_option("mode").and_then(|it| it.parse().ok())
+    }
+
+    pub fn pretty(&self) -> Pretty {
+        self.get_option("pretty")
+            .and_then(|it| it.parse().ok())
+            .unwrap_or_default()
+    }
+
+    pub fn date(&self) -> String {
+        self.get_option("date")
+            .unwrap_or_else(|| "%Y%m%d".to_string())
+    }
+
+    pub fn gitexport(&self) -> GitExport {
+        self.get_option("gitexport")
+            .and_then(|it| it.parse().ok())
+            .unwrap_or_default()
+    }
+
+    pub fn gitmode(&self) -> GitMode {
+        self.get_option("gitmode")
+            .and_then(|it| it.parse().ok())
+            .unwrap_or_default()
+    }
+
+    pub fn pgpmode(&self) -> PgpMode {
+        self.get_option("pgpmode")
+            .and_then(|it| it.parse().ok())
+            .unwrap_or_default()
+    }
+
+    pub fn searchmode(&self) -> SearchMode {
+        self.get_option("searchmode")
+            .and_then(|it| it.parse().ok())
+            .unwrap_or_default()
+    }
+
+    pub fn decompress(&self) -> bool {
+        self.has_option("decompress")
+    }
+
+    pub fn bare(&self) -> bool {
+        self.has_option("bare")
+    }
+
+    pub fn user_agent(&self) -> Option<String> {
+        self.get_option("user-agent")
+    }
+
+    pub fn passive(&self) -> Option<bool> {
+        if self.has_option("passive") || self.has_option("pasv") {
+            Some(true)
+        } else if self.has_option("active") || self.has_option("nopasv") {
+            Some(false)
+        } else {
+            None
+        }
+    }
+
+    pub fn unzipoptions(&self) -> Option<String> {
+        self.get_option("unzipopt")
+    }
+
+    pub fn dversionmangle(&self) -> Option<String> {
+        self.get_option("dversionmangle")
+    }
+
+    pub fn dirversionmangle(&self) -> Option<String> {
+        self.get_option("dirversionmangle")
+    }
+
+    pub fn pagemangle(&self) -> Option<String> {
+        self.get_option("pagemangle")
+    }
+
+    pub fn uversionmangle(&self) -> Option<String> {
+        self.get_option("uversionmangle")
+    }
+
+    pub fn versionmangle(&self) -> Option<String> {
+        self.get_option("versionmangle")
+    }
+
+    pub fn hrefdecode(&self) -> bool {
+        self.get_option("hrefdecode").is_some()
+    }
+
+    pub fn downloadurlmangle(&self) -> Option<String> {
+        self.get_option("downloadurlmangle")
+    }
+
+    pub fn filenamemangle(&self) -> Option<String> {
+        self.get_option("filenamemangle")
+    }
+
+    pub fn pgpsigurlmangle(&self) -> Option<String> {
+        self.get_option("pgpsigurlmangle")
+    }
+
+    pub fn oversionmangle(&self) -> Option<String> {
+        self.get_option("oversionmangle")
+    }
+
     /// Returns options set
     pub fn opts(&self) -> std::collections::HashMap<String, String> {
         let mut options = std::collections::HashMap::new();
@@ -435,19 +602,87 @@ impl Entry {
         self.items().nth(1)
     }
 
-    pub fn version(&self) -> Option<String> {
-        self.items().nth(2)
+    /// Returns the version policy
+    pub fn version(&self) -> Option<crate::VersionPolicy> {
+        self.items().nth(2).and_then(|it| it.parse().ok())
     }
 
+    /// Returns the script of the entry.
     pub fn script(&self) -> Option<String> {
         self.items().nth(3)
     }
+
+    /// Replace all substitutions and return the resulting URL.
+    pub fn format_url(&self, package: impl FnOnce() -> String) -> url::Url {
+        subst(self.url().as_str(), package).parse().unwrap()
+    }
+}
+
+const SUBSTITUTIONS: &[(&str, &str)] = &[
+    // This is substituted with the source package name found in the first line
+    // of the debian/changelog file.
+    // "@PACKAGE@": None,
+    // This is substituted by the legal upstream version regex (capturing).
+    ("@ANY_VERSION@", r"[-_]?(\d[\-+\.:\~\da-zA-Z]*)"),
+    // This is substituted by the typical archive file extension regex
+    // (non-capturing).
+    (
+        "@ARCHIVE_EXT@",
+        r"(?i)\.(?:tar\.xz|tar\.bz2|tar\.gz|zip|tgz|tbz|txz)",
+    ),
+    // This is substituted by the typical signature file extension regex
+    // (non-capturing).
+    (
+        "@SIGNATURE_EXT@",
+        r"(?i)\.(?:tar\.xz|tar\.bz2|tar\.gz|zip|tgz|tbz|txz)\.(?:asc|pgp|gpg|sig|sign)",
+    ),
+    // This is substituted by the typical Debian extension regexp (capturing).
+    ("@DEB_EXT@", r"[\+~](debian|dfsg|ds|deb)(\.)?(\d+)?$"),
+];
+
+pub fn subst(text: &str, package: impl FnOnce() -> String) -> String {
+    let mut substs = SUBSTITUTIONS.to_vec();
+    let package_name;
+    if text.contains("@PACKAGE@") {
+        package_name = Some(package());
+        substs.push(("@PACKAGE@", package_name.as_deref().unwrap()));
+    }
+
+    let mut text = text.to_string();
+
+    for (k, v) in substs {
+        text = text.replace(k, v);
+    }
+
+    text
+}
+
+#[test]
+fn test_subst() {
+    assert_eq!(
+        subst("@ANY_VERSION@", || unreachable!()),
+        r"[-_]?(\d[\-+\.:\~\da-zA-Z]*)"
+    );
+    assert_eq!(subst("@PACKAGE@", || "dulwich".to_string()), "dulwich");
 }
 
 impl OptionList {
     fn children(&self) -> impl Iterator<Item = _Option> + '_ {
         self.0.children().filter_map(_Option::cast)
     }
+
+    pub fn has_option(&self, key: &str) -> bool {
+        self.children().any(|it| it.key().as_deref() == Some(key))
+    }
+
+    pub fn get_option(&self, key: &str) -> Option<String> {
+        for child in self.children() {
+            if child.key().as_deref() == Some(key) {
+                return child.value();
+            }
+        }
+        None
+    }
 }
 
 impl _Option {
@@ -486,35 +721,38 @@ impl _Option {
 #[test]
 fn test_parse_v1() {
     const WATCHV1: &str = r#"version=4
-opts=filenamemangle=s/.+\/v?(\d\S+)\.tar\.gz/syncthing-gtk-$1\.tar\.gz/ \
+opts=bare,filenamemangle=s/.+\/v?(\d\S+)\.tar\.gz/syncthing-gtk-$1\.tar\.gz/ \
   https://github.com/syncthing/syncthing-gtk/tags .*/v?(\d\S+)\.tar\.gz
 "#;
     let parsed = parse(WATCHV1);
-    assert_eq!(parsed.errors, Vec::<String>::new());
+    //assert_eq!(parsed.errors, Vec::<String>::new());
     let node = parsed.syntax();
     assert_eq!(
         format!("{:#?}", node),
-        r#"ROOT@0..156
+        r#"ROOT@0..161
   VERSION@0..10
     KEY@0..7 "version"
     EQUALS@7..8 "="
     VALUE@8..9 "4"
     NEWLINE@9..10 "\n"
-  ENTRY@10..156
-    OPTS_LIST@10..81
+  ENTRY@10..161
+    OPTS_LIST@10..86
       KEY@10..14 "opts"
       EQUALS@14..15 "="
-      OPTION@15..81
-        KEY@15..29 "filenamemangle"
-        EQUALS@29..30 "="
-        VALUE@30..81 "s/.+\\/v?(\\d\\S+)\\.tar\\ ..."
-    WHITESPACE@81..82 " "
-    CONTINUATION@82..84 "\\\n"
-    WHITESPACE@84..86 "  "
-    VALUE@86..133 "https://github.com/sy ..."
-    WHITESPACE@133..134 " "
-    VALUE@134..155 ".*/v?(\\d\\S+)\\.tar\\.gz"
-    NEWLINE@155..156 "\n"
+      OPTION@15..19
+        KEY@15..19 "bare"
+      COMMA@19..20 ","
+      OPTION@20..86
+        KEY@20..34 "filenamemangle"
+        EQUALS@34..35 "="
+        VALUE@35..86 "s/.+\\/v?(\\d\\S+)\\.tar\\ ..."
+    WHITESPACE@86..87 " "
+    CONTINUATION@87..89 "\\\n"
+    WHITESPACE@89..91 "  "
+    VALUE@91..138 "https://github.com/sy ..."
+    WHITESPACE@138..139 " "
+    VALUE@139..160 ".*/v?(\\d\\S+)\\.tar\\.gz"
+    NEWLINE@160..161 "\n"
 "#
     );
 
@@ -574,4 +812,121 @@ https://github.com/syncthing/syncthing-gtk/tags .*/v?(\d\S+)\.tar\.gz
         entry.url(),
         "https://github.com/syncthing/syncthing-gtk/tags"
     );
+    assert_eq!(
+        entry.format_url(|| "syncthing-gtk".to_string()),
+        "https://github.com/syncthing/syncthing-gtk/tags"
+            .parse()
+            .unwrap()
+    );
+}
+
+#[test]
+fn test_parse_v3() {
+    let parsed = parse(
+        r#"version=4
+https://github.com/syncthing/@PACKAGE@/tags .*/v?(\d\S+)\.tar\.gz
+# comment
+"#,
+    );
+    assert_eq!(parsed.errors, Vec::<String>::new());
+    let root = parsed.root();
+    assert_eq!(root.version(), 4);
+    let entries = root.entries().collect::<Vec<_>>();
+    assert_eq!(entries.len(), 1);
+    let entry = &entries[0];
+    assert_eq!(entry.url(), "https://github.com/syncthing/@PACKAGE@/tags");
+    assert_eq!(
+        entry.format_url(|| "syncthing-gtk".to_string()),
+        "https://github.com/syncthing/syncthing-gtk/tags"
+            .parse()
+            .unwrap()
+    );
+}
+
+#[test]
+fn test_parse_v4() {
+    let cl: super::WatchFile = r#"version=4
+opts=repack,compression=xz,dversionmangle=s/\+ds//,repacksuffix=+ds \
+    https://github.com/example/example-cat/tags \
+        (?:.*?/)?v?(\d[\d.]*)\.tar\.gz debian uupdate
+"#
+    .parse()
+    .unwrap();
+    assert_eq!(cl.version(), 4);
+    let entries = cl.entries().collect::<Vec<_>>();
+    assert_eq!(entries.len(), 1);
+    let entry = &entries[0];
+    assert_eq!(entry.url(), "https://github.com/example/example-cat/tags");
+    assert_eq!(
+        entry.matching_pattern(),
+        Some("(?:.*?/)?v?(\\d[\\d.]*)\\.tar\\.gz".into())
+    );
+    assert!(entry.repack());
+    assert_eq!(entry.compression(), Some(Compression::Xz));
+    assert_eq!(entry.dversionmangle(), Some("s/\\+ds//".into()));
+    assert_eq!(entry.repacksuffix(), Some("+ds".into()));
+    assert_eq!(entry.script(), Some("uupdate".into()));
+    assert_eq!(
+        entry.format_url(|| "example-cat".to_string()),
+        "https://github.com/example/example-cat/tags"
+            .parse()
+            .unwrap()
+    );
+    assert_eq!(entry.version(), Some(VersionPolicy::Debian));
+}
+
+#[test]
+fn test_git_mode() {
+    let text = r#"version=3
+opts="mode=git, gitmode=shallow, pgpmode=gittag" \
+https://git.kernel.org/pub/scm/linux/kernel/git/firmware/linux-firmware.git \
+refs/tags/(.*) debian
+"#;
+    let parsed = parse(text);
+    assert_eq!(parsed.errors, Vec::<String>::new());
+    let cl = parsed.root();
+    assert_eq!(cl.version(), 3);
+    let entries = cl.entries().collect::<Vec<_>>();
+    assert_eq!(entries.len(), 1);
+    let entry = &entries[0];
+    assert_eq!(
+        entry.url(),
+        "https://git.kernel.org/pub/scm/linux/kernel/git/firmware/linux-firmware.git"
+    );
+    assert_eq!(entry.matching_pattern(), Some("refs/tags/(.*)".into()));
+    assert_eq!(entry.version(), Some(VersionPolicy::Debian));
+    assert_eq!(entry.script(), None);
+    assert_eq!(entry.gitmode(), GitMode::Shallow);
+    assert_eq!(entry.pgpmode(), PgpMode::GitTag);
+    assert_eq!(entry.mode(), Some(Mode::Git));
+}
+
+#[test]
+fn test_parse_quoted() {
+    const WATCHV1: &str = r#"version=4
+opts="bare, filenamemangle=blah" \
+  https://github.com/syncthing/syncthing-gtk/tags .*/v?(\d\S+)\.tar\.gz
+"#;
+    let parsed = parse(WATCHV1);
+    //assert_eq!(parsed.errors, Vec::<String>::new());
+    let node = parsed.syntax();
+
+    let root = parsed.root();
+    assert_eq!(root.version(), 4);
+    let entries = root.entries().collect::<Vec<_>>();
+    assert_eq!(entries.len(), 1);
+    let entry = &entries[0];
+
+    assert_eq!(
+        entry.url(),
+        "https://github.com/syncthing/syncthing-gtk/tags"
+    );
+    assert_eq!(
+        entry.matching_pattern(),
+        Some(".*/v?(\\d\\S+)\\.tar\\.gz".into())
+    );
+    assert_eq!(entry.version(), None);
+    assert_eq!(entry.script(), None);
+
+    assert_eq!(node.text(), WATCHV1);
 }
